module AlsaCap
  # This class signals that an error has occurred during the interaction with
  # ALSA.
  class ALSAError < StandardError
  end

  private

  # Throws an exception with an informative message.
  #
  # * *Args*    :
  #   - +errno+ -> error number.
  #   - +what+ -> description of the current action when the error happened.
  #   - +card+ -> card number.
  #   - +dev+ -> device number, if applicable.
  # * *Raises* :
  #   - +ALSAError+ -> in all cases.
  def self.throw_AlsaError(errno, what, card, dev = nil)
    if dev
      fail ALSAError.new("Error #{what} for card #{card}, device #{dev}: "\
                         "#{self.err_to_s(errno)}. Skipping.")
    else
      fail ALSAError.new("Error #{what} for card #{card}: "\
                         "#{self.err_to_s(errno)}. Skipping.")
    end
  end

  # Returns a string describing error number.
  #
  # * *Args*    :
  #   - +err+ -> error number.
  # * *Returns* :
  #   - a string describing error number.
  def self.err_to_s(err)
    case err.abs()
    when 0 then 'No error'
    when Errno::EBADFD::Errno   then 'PCM device is in a bad state'
    when Errno::EPIPE::Errno    then 'An underrun occurred'
    when Errno::ESTRPIPE::Errno then 'A suspend event occurred'
    when Errno::ENOTTY::Errno   then 'Hotplug device has been removed'
    when Errno::ENODEV::Errno   then 'Hotplug device has been removed'
    when Errno::ENOENT::Errno   then 'Device does not exist'
    else SystemCallError.new('', err.abs()).message.gsub(/ - .*/, '')
    end
  end
end
