require 'rb-alsacap/libasound'
require 'rb-alsacap/alsa_error'
require 'rb-alsacap/device'

module AlsaCap
  # A class that represents an audio card.
  class Card
    attr_reader :number
    attr_reader :id
    attr_reader :name

    # Returns a new card.
    #
    # * *Args*    :
    #   - +card_n+ -> card number.
    #   - +stream+ -> PCM stream direction.
    def initialize(number, stream)
      @number  = number
      # Get the handle.
      ctl_handle = get_handle()
      # Retrieve information about the card.
      info = FFI::MemoryPointer.new(LibAsound.snd_ctl_card_info_sizeof())
      if LibAsound.snd_ctl_card_info(ctl_handle, info) < 0
        AlsaCap.throw_AlsaError(err, 'obtaining card info', @number)
      end
      @id      = LibAsound.snd_ctl_card_info_get_id(info)
      @name    = LibAsound.snd_ctl_card_info_get_name(info)
      @devices = []
      # Process devices.
      dev_n = FFI::MemoryPointer.new(:int)
      dev_n.write_int(-1)
      err = LibAsound.snd_ctl_pcm_next_device(ctl_handle, dev_n)
      while dev_n.read_int() >= 0
        begin
          @devices << Device.new(@number, dev_n.read_int(), ctl_handle, stream)
        rescue ALSAError => err
          @devices << err.message
        end
        err = LibAsound.snd_ctl_pcm_next_device(ctl_handle, dev_n)
      end
    ensure
      LibAsound.snd_ctl_close(ctl_handle) if ctl_handle != FFI::Pointer::NULL
    end

    # Returns a string representing the card.
    #
    # * *Returns* :
    #   - a string representing the card.
    def to_s
      str = "Card #{@number}, ID '#{@id}', name '#{@name}'\n"
      @devices.each() { |dev| str << '  ' << dev.to_s() << "\n" }
      str.chomp!()
    end

    private

    # Returns a CTL handle on the card.
    #
    # * *Returns* :
    #   - a CTL handle on the card.
    # * *Raises* :
    #   - +ALSAError+ -> if we cannot obtain the handle.
    def get_handle()
      hwdev = "hw:#{@number}"
      # Wrap the 'pcm' pointer into a temporary pointer.
      ctl_handle = FFI::Pointer::NULL
      ref_handle = FFI::MemoryPointer.new(:pointer)
      ref_handle.write_pointer(ctl_handle)
      if LibAsound.snd_ctl_open(ref_handle, hwdev, 0) < 0
        AlsaCap.throw_AlsaError(err, 'obtaining control interface', @number)
      end
      ref_handle.read_pointer()
    end
  end
end
