require 'rb-alsacap/libasound'
require 'rb-alsacap/alsa_error'
require 'rb-alsacap/subdevice'

module AlsaCap
  # A class that represents an audio device.
  class Device
    attr_reader :number
    attr_reader :id
    attr_reader :name

    # Returns a new device.
    #
    # * *Args*    :
    #   - +card_n+ -> card number.
    #   - +dev_n+ -> device number.
    #   - +ctl_handle+ -> a CTL handle.
    #   - +stream+ -> PCM stream direction.
    def initialize(card_n, dev_n, ctl_handle, stream)
      @number = dev_n
      # Retrieve information.
      pcm_info = FFI::MemoryPointer.new(LibAsound.snd_pcm_info_sizeof())
      setup_device(card_n, pcm_info, ctl_handle, stream)
      @id    = LibAsound.snd_pcm_info_get_id(pcm_info)
      @name  = LibAsound.snd_pcm_info_get_name(pcm_info)
      @avail = LibAsound.snd_pcm_info_get_subdevices_avail(pcm_info)
      # Retrieve the number of subdevices.
      nb_subdev = get_subdevice(card_n, pcm_info)
      # Retrieve the hardware parameters.
      hw_params = get_hw_params(card_n, stream)
      # Get channels.
      uint_ptr = FFI::MemoryPointer.new(:uint, 1)
      LibAsound.snd_pcm_hw_params_get_channels_min(hw_params, uint_ptr)
      @min_channel = uint_ptr.read_uint()
      LibAsound.snd_pcm_hw_params_get_channels_max(hw_params, uint_ptr)
      @max_channel = uint_ptr.read_uint()
      # Get sampling rates.
      LibAsound.snd_pcm_hw_params_get_rate_min(hw_params, uint_ptr, FFI::Pointer::NULL)
      @min_rate= uint_ptr.read_uint()
      LibAsound.snd_pcm_hw_params_get_rate_max(hw_params, uint_ptr, FFI::Pointer::NULL)
      @max_rate = uint_ptr.read_uint()
      # Get formats.
      @format_mask = FFI::MemoryPointer.new(LibAsound.snd_pcm_format_mask_sizeof())
      LibAsound.snd_pcm_hw_params_get_format_mask(hw_params, @format_mask)
      # Process subdevices.
      @subdevices  = []
      (0..nb_subdev-1).each do |subdev|
        begin
          @subdevices << SubDevice.new(card_n, dev_n, subdev, pcm_info, ctl_handle)
        rescue ALSAError => err
          @subdevices << err.message
        end
      end
    end

    # Returns a string representing the device.
    #
    # * *Returns* :
    #   - a string representing the device.
    def to_s
      str = "Device #{@number}, ID '#{@id}', name '#{@name}', "
      str << "#{@subdevices.length} subdevices (#{@avail} available)\n"
      str << channel_to_s() << "sampling rate #{@min_rate}..#{@max_rate} Hz\n"
      str << '    Sample formats: ' << format_mask_to_s() << "\n"
      @subdevices.each() { |subdev| str << '      ' << subdev.to_s() << "\n" }
      str.chomp!()
    end

    private

    # Setups the device and retrieve information about it.
    #
    # * *Args*    :
    #   - +card_n+ -> card number.
    #   - +pcm_info+ -> a PCM generic info structure.
    #   - +ctl_handle+ -> a CTL handle.
    #   - +stream+ -> PCM stream direction.
    # * *Raises* :
    #   - +ALSAError+ -> if we cannot obtain the information about the device.
    def setup_device(card_n, pcm_info, ctl_handle, stream)
      LibAsound.snd_pcm_info_set_device(pcm_info, @number)
      LibAsound.snd_pcm_info_set_subdevice(pcm_info, 0)
      LibAsound.snd_pcm_info_set_stream(pcm_info, stream)
      err = LibAsound.snd_ctl_pcm_info(ctl_handle, pcm_info)
      if err < 0
        AlsaCap.throw_AlsaError(err, 'obtaining device info', card_n, @number)
      end
    end

    # Returns the number of subdevices for this device.
    #
    # * *Args*    :
    #   - +card_n+ -> card number.
    #   - +pcm_info+ -> a PCM generic info structure.
    # * *Returns* :
    #   - the numer of subdevices for this device.
    # * *Raises* :
    #   - +ALSAError+ -> if we cannot obtain the number of subdevices.
    def get_subdevice(card_n, pcm_info)
      nb_subdev = LibAsound.snd_pcm_info_get_subdevices_count(pcm_info)
      if nb_subdev < 0
        AlsaCap.throw_AlsaError(err, 'obtaining device info', card_n, @number)
      end
      return nb_subdev
    end

    # Returns the hardware parameters of the device.
    #
    # * *Args*    :
    #   - +card_n+ -> card number.
    #   - +stream+ -> PCM stream direction.
    # * *Returns* :
    #   - the hardware parameters of the device.
    # * *Raises* :
    #   - +ALSAError+ -> if we cannot obtain the information.
    def get_hw_params(card_n, stream)
      # Wrap the 'pcm' pointer into a temporary pointer.
      pcm = FFI::Pointer::NULL
      ref_pcm = FFI::MemoryPointer.new(:pointer)
      ref_pcm.write_pointer(pcm)
      # Open the device.
      hwdev = "hw:#{card_n},#{@number}"
      err = LibAsound.snd_pcm_open(ref_pcm, hwdev, stream, LibAsound::SND_PCM_NONBLOCK)
      if err < 0
        AlsaCap.throw_AlsaError(err, 'obtaining sound device', card_n, @number)
      end
      pcm = ref_pcm.read_pointer()
      # Retrieve the parameters.
      hw_params = FFI::MemoryPointer.new(LibAsound.snd_pcm_hw_params_sizeof())
      err = LibAsound.snd_pcm_hw_params_any(pcm, hw_params)
      LibAsound.snd_pcm_close(pcm)
      if err < 0
        AlsaCap.throw_AlsaError(err, 'obtaining hardware parameters', card_n, @number)
      end
      return hw_params
    end

    # Returns a string representing the list of supported format.
    #
    # * *Returns* :
    #   - a string representing the list of supported format.
    def channel_to_s
      if @min_channel == @max_channel
        if @min_channel == 1
          '    1 channel, '
        else
          "    #{@min_channel} channels, "
        end
      else
        "    #{@min_channel}..#{@max_channel} channels, "
      end
    end

    # Returns a string representing the list of supported format.
    #
    # * *Returns* :
    #   - a string representing the list of supported format.
    def format_mask_to_s
      formats = []
      last = LibAsound.enum_type(:snd_pcm_format_t)[:SND_PCM_FORMAT_LAST]
      (0 .. last).each do |format|
        if LibAsound.snd_pcm_format_mask_test(@format_mask, format) != 0
          formats << LibAsound.snd_pcm_format_name(format)
        end
      end

      if formats.size == 0
        '(none)'
      else
        formats.join(', ')
      end
    end
  end
end
